//---------------------------------------------------------------------------

#include "FonctionQuantile.h"

//---------------------------------------------------------------------------

//#include <conio.h>
#include <math.h>
#include <float.h>
#define pi 3.14159265358979

//---------------------------------------------------------------------------


/**
 * Fonction d'erreur inverse.
 * Source : Matlab
 * Utilise 2 pas de correction de Newton-Raphson.
 * @param y de type double.
 * @see double Erf(double x);
 * @return Valeur de la fonction d'erreur inverse en y.
 */
double erfinv(double y)
{
double Erf(double x);
static double a[4] = {0.886226899, -1.645349621, 0.914624893, -0.140543331};
static double b[4] = {-2.118377725, 1.442710462, -0.329097515, 0.012229801};
static double c[4] = {-1.970840454, -1.624906493, 3.429567803, 1.641345311};
static double d[2] = {3.543889200, 1.637067800};
double y0,z,x,temp;

	y0 = 0.7;
	if (fabs(y)<=y0)
	{
		z = y * y;
		x = y * (((a[3] * z + a[2]) * z + a[1]) * z + a[0]) / ((((b[3] * z + b[2]) * z + b[1]) * z + b[0]) * z + 1);
	}
	if ((y>y0) && (y<1))
	{
		z = sqrt(-log((1-y)/2.0));
		x = (((c[3] * z + c[2]) * z + c[1]) * z + c[0]) / ((d[1] * z + d[0]) * z + 1);
	}
	if ((y<(-y0)) && (y>(-1)))
	{
		z = sqrt(-log((1+y)/2.0));
		x = -(((c[3] * z + c[2]) * z + c[1]) * z + c[0]) / ((d[1] * z + d[0]) * z + 1);
	}
	
	if (y==-1)
		return(-DBL_MAX);
	if (y==1)
		return(DBL_MAX);
	if(fabs(y)>1)
		return(0);

	temp = x - (Erf(x) - y) / (2/sqrt(pi) * exp(-(x*x)));
	return(temp - (Erf(temp) - y) / (2/sqrt(pi) * exp(-(temp*temp))));
}

// IMPLEMENTATION DES METHODES

// Constructeur
FonctionQuantile::FonctionQuantile()
{
}

FonctionQuantile::~FonctionQuantile()
{
}

double FonctionQuantile::QuantileExponentielle(double y, double alpha)
{
	if ((y<0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
			return(-alpha*log(1-y));
	}
}

double FonctionQuantile::QuantileWeibull(double y, double a, double b)
{
	return(pow(QuantileExponentielle(y,a),1/b));
}

double FonctionQuantile::QuantilePareto(double y,double x0, double alpha)
{
double temp;
	if ((y<0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
		{
			temp = pow(1-y,1/alpha);
			return(x0/temp);
		}
	}
}

double FonctionQuantile::QuantileGPD(double y, double gam, double sigma)
{
double temp,constante,x;

	if ((y<0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
		{
			if (gam!=0)
			{
				constante = sigma/gam;
				temp = pow(1-y,gam);
				x = constante/temp - constante;
				if ((x>0) || ((x>=0) && (x<(-sigma/gam))))
					return(x);
				else
					return(0);
			}
			else
				return(QuantileExponentielle(y,sigma));
		}
	}
}

double FonctionQuantile::QuantileNormale(double y, double a, double b)
{
	if ((y<0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
		{
			if (y==0)
				return(-DBL_MAX);
			else
				return(sqrt(2.0) * b * erfinv(2.0 * y - 1) + a);
		}
	}
}

double FonctionQuantile::QuantileLognormale(double y, double a, double b)
{
	if ((y<=0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
			return(exp(sqrt(2) * b * erfinv(2 * y - 1) + a));
	}
}

double FonctionQuantile::QuantileGamma(double y, double a, double teta)
{
double mn,temp,sigma,v,mu,xk,h,xnew;
int count, count_limit;
const double eps = 2.22e-16;
FonctionRepartition F;
Densite D;

	if ((y<=0) || (y>1))
		return(0);
	else
	{
		if (y==1)
			return(DBL_MAX);
		else
		{
			count_limit = 100;
			count = 0;
			mn = a * teta;
			v = mn * teta;
			temp = log(v + mn*mn); 
			mu = 2 * log(mn) - 0.5 * temp;
			sigma = -2 * log(mn) + temp;
			xk = exp(QuantileNormale(y,mu,sigma));

			h = 1;
			while( (fabs(h)>sqrt(eps)*fabs(xk))  &&  (fabs(h)>sqrt(eps))  && (count < count_limit) )
			{
				count = count + 1;
				h = (F.FonctionGamma(xk,a,teta) - y) / D.DensiteGamma(xk,a,teta);
				xnew = xk - h;
				
				if (xnew<0)
				{
					xnew = xk / 10.0;
					h = xk - xnew;
				}
				xk = xnew;
			}
			if (count == count_limit)
				return(-10);
			else
				return(xk);
		}
	}
}

double FonctionQuantile::QuantileBeta1espece(double y, double a, double b)
{
int CountLimit,Count;
double eps = 2.22e-16;
FonctionRepartition F;
Densite D;
double xk,xnew,h,crit;

	if (y <= 0)
		return (0);
	else
	if (y == 1)
		return (1);
	if (y > 1)
		return (0);
	CountLimit = 100;
	Count = 0;
	xk = a / (a + b);
	if (xk == 0)
		xk = sqrt(eps);
	if (xk == 1)
		xk = 1 - sqrt(eps);
	h = 1;
	crit = sqrt(eps);
	while((fabs(h) > crit * fabs(xk)) && (fabs(h) > crit) && (Count < CountLimit))
	{
		Count = Count + 1;
		h = (F.FonctionBeta1espece(xk,a,b) - y) / D.DensiteBeta1espece(xk,a,b);
		xnew = xk - h;
		if (xnew <= 0)
			xnew = xk / 10.0;
		if (xnew >= 1)
			xnew = 1 - (1 - xk) / 10.0;
		xk = xnew;
	}
	if (Count == CountLimit)
		return (2);
	else
		return(xk);
}

//**********************************************************************************

double FonctionQuantile::QuantileChi2(double y, int n)
{
	return(QuantileGamma(y,n/2.0,2));
}

//**********************************************************************************

double FonctionQuantile::QuantileStudent(double y, int n)
{
double Temp,df,xn;

	if (n == 1)
		return(tan(pi * (y - 0.5)));
	else
	{
		if (n < 1000)
		{
			if ((y >= 0.5) && (y < 1))
			{
				Temp = QuantileBeta1espece(2 * (1 - y),n / 2.0,0.5);
				return(sqrt(n / Temp - n));
			}
			else
			{
				Temp = QuantileBeta1espece(2 * y,n / 2.0,0.5);
				return(- sqrt(n / Temp - n));
			}
		}
		else
		{
			xn = QuantileNormale(y,0,1);
			df = (double)n;
			Temp = xn + (pow(xn,3) + xn) / (4 * df) + (5 * pow(xn,5) + 16 * pow(xn,3) + 3 * xn) / (96 * df * df) +
					(3 * pow(xn,7) + 19 * pow(xn,5) + 17 * pow(xn,3) - 15 * xn) / (384 * pow(df,3)) +
					(79 * pow(xn,9) + 776 * pow(xn,7) + 1482 * pow(xn,5) - 1920 * pow(xn,3) - 945 * xn) / (92160 * pow(df,4));
			return(Temp);
		}
	}
}

//**********************************************************************************

double FonctionQuantile::QuantileUniforme(double y, double a, double b)
{
double Temp;

	if ((y<0) || (y>1))
		return(0);
	else
	{
		Temp = a + y * (b - a);
		return(Temp);
	}
}


//**********************************************************************************

double FonctionQuantile::QuantileCalcul(Loi Nomloi, Parametre &P, double y)
{
	switch(Nomloi)
	{
	case Exponentielle : return(QuantileExponentielle(y,P.element(1)));
	case Weibull : return(QuantileWeibull(y,P.element(1),P.element(2)));
	case Pareto : return(QuantilePareto(y,P.element(1),P.element(2)));
	case GPD : return(QuantileGPD(y,P.element(1),P.element(2)));
	case Normale : return(QuantileNormale(y,P.element(1),P.element(2)));
	case Lognormale : return(QuantileLognormale(y,P.element(1),P.element(2)));
	case LoiGamma : return(QuantileGamma(y,P.element(1),P.element(2)));
	case Beta1espece : return(QuantileBeta1espece(y,P.element(1),P.element(2)));
	case Chi2 : return(QuantileChi2(y,(int)P.element(1)));
	case Student : return(QuantileStudent(y,(int)P.element(1)));
	case UniformeContinue : return(QuantileUniforme(y,P.element(1),P.element(2)));
	default : return 0;
	}
}







